complete -c grimpicker -f
complete -c grimpicker -s p -l print -d "Print to stdout"
complete -c grimpicker -s d -l draw -d "Draw a colored block"
complete -c grimpicker -s e -l escape -d "Print shell escape sequences"
complete -c grimpicker -s c -l copy -d "Copy to clipboard"
complete -c grimpicker -s n -l notify -d "Send a notification"
complete -c grimpicker -s h -l help -d "Show help message and quit"
complete -c grimpicker -s v -l version -d "Show version number and quit"
